/*
 *	fft.c
 *
 *	C Version 1.6 by Steve Sampson, Public Domain, 15 August 1988
 *
 *	This program is based on the work by W. D. Stanley and S. J. Peterson,
 *	Old Dominion University.  It produces a Frequency Domain display from
 *	the Time Domain data input using the Fast Fourier Transform.
 *
 *	The Real data is generated by the in-phase (I) channel and the
 *	Imaginary data is produced by the quadrature-phase (Q) channel of
 *	a Doppler Radar receiver.  The middle filter is zero Hz.  Closing
 *	targets are displayed to the right, and Opening targets to the left.
 *
 *	Note: With Imaginary data set to zero the output is a mirror image.
 *
 *	Usage:	fft samples input output
 *		1. Samples is a power of two.
 *		2. Input is (samples * sizeof(double)) characters.
 *		3. Output is (samples * sizeof(double)) characters.
 *		4. Standard error is help or debugging messages.
 *
 *	Define SPEED to produce a program that pre-computes the sin/cos tables,
 *		this takes more memory however.
 *
 *	Define SIMULATE to substitute a quicky magnitude routine, Has a max
 *		12% error.
 *
 *	Define SYSV to create a Unix(tm) System V operating system version,
 *		else it defaults to MS-DOS(tm) and Turbo-C(tm) V1.5
 *		and MS-C V5.0.
 *
 *	See also: readme.doc, gen_pulse.c, and gen_sine.c.
 */

/* Includes */

#include <stdio.h>
#include <malloc.h>	/* Rename Turbo-C alloc.h to malloc.h */
#include <math.h>


/* Defines */

#define	TWO_PI	((double)2.0 * 3.14159265358979324)
#define	Chunk	((unsigned)(samples * sizeof(double)))

#ifdef	SYSV
#define	RMODE	"r"
#define	WMODE	"w"
#else
#define	RMODE	"rb"
#define	WMODE	"wb"
#endif


/* Globals, Forward declarations */

static int	samples, power, permute();
static double	*real, *imag, maxn, magnitude();
static void	fft(), max_amp(), display(), err_report();

#ifdef SPEED
static double	*sine, *cosine;
static void	build_trig();
#endif

static FILE	*fpi, *fpo;


/* The program */

main(argc, argv)
int	argc;
char	*argv[];
{
	if (argc != 4)
		err_report(0);

	samples = abs(atoi(argv[1]));
	power   = (int)(log10((double)samples) / log10((double)2.0));

	/* Allocate memory for the variable arrays */

	if ((real = (double *)malloc(Chunk)) == (double *)NULL)
		err_report(1);

	if ((imag = (double *)malloc(Chunk)) == (double *)NULL)
		err_report(2);

#ifdef SPEED
	if ((sine = (double *)malloc(Chunk)) == (double *)NULL)
		err_report(3);

	if ((cosine = (double *)malloc(Chunk)) == (double *)NULL)
		err_report(4);
#endif
	/* open the data files */

	if ((fpi = fopen(argv[2], RMODE)) == (FILE *)NULL)
		err_report(5);

	if ((fpo = fopen(argv[3], WMODE)) == (FILE *)NULL)
		err_report(6);

	/* read in the data from the input file */

	fread((char *)real, sizeof(double), (unsigned)samples, fpi);
	fread((char *)imag, sizeof(double), (unsigned)samples, fpi);
	fclose(fpi);

#ifdef SPEED
	build_trig();
#endif
	fft();
	display();

	/* write the frequency domain data to the standard output */

	fwrite((char *)real, sizeof(double), (unsigned)samples, fpo);
	fwrite((char *)imag, sizeof(double), (unsigned)samples, fpo);
	fclose(fpo);

	free((char *)real);
	free((char *)imag);

#ifdef SPEED
	free((char *)sine);
	free((char *)cosine);
#endif

	exit(0);
}


static void err_report(n)
int	n;
{
	switch (n)  {
	case 0:
		fprintf(stderr, "Usage: fft samples\n");
		fprintf(stderr, "Where samples is a power of two\n");
		break;
	case 1:
		fprintf(stderr, "fft: Out of memory getting real space\n");
		break;
	case 2:
		fprintf(stderr, "fft: Out of memory getting imag space\n");
		free((char *)real);
		break;
#ifdef SPEED
	case 3:
		fprintf(stderr, "fft: Out of memory getting sine space\n");
		free((char *)real);
		free((char *)imag);
		break;
	case 4:
		fprintf(stderr, "fft: Out of memory getting cosine space\n");
		free((char *)real);
		free((char *)imag);
		free((char *)sine);
		break;
#endif
	case 5:
		fprintf(stderr,"fft: Unable to open data input file\n");
		free((char *)real);
		free((char *)imag);
#ifdef SPEED
		free((char *)sine);
		free((char *)cosine);
#endif
		break;
	case 6:
		fprintf(stderr,"fft: Unable to open data output file\n");
		fclose(fpi);
		free((char *)real);
		free((char *)imag);
#ifdef SPEED
		free((char *)sine);
		free((char *)cosine);
#endif
	}

	exit(1);
}


static void fft()
{
	register int	loop, loop1, loop2;
	unsigned	i1;			/* going to right shift this */
	int		i2, i3, i4, y;
	double		a1, a2, b1, b2, z1, z2;
#ifndef SPEED
	double		v;

	v  = TWO_PI * ((double)1.0 / (double)samples);
#endif
	i1 = samples / 2;
	i2 = 1;

	/* perform the butterflies */

	for (loop = 0; loop < power; loop++)  {
		i3 = 0;
		i4 = i1;

		for (loop1 = 0; loop1 < i2; loop1++)  {
			y = permute(i3 / (int)i1);
#ifdef SPEED
			z1 =  cosine[y];
			z2 = -sine[y];
#else
			z1 =  cos(v * (double)y);
			z2 = -sin(v * (double)y);
#endif
			for (loop2 = i3; loop2 < i4; loop2++)  {

				/* Don't depend on complex precedences */

				a1 = real[loop2];
				a2 = imag[loop2];

				b1  = z1 * real[loop2 + i1];
				b1 -= z2 * imag[loop2 + i1];

				b2  = z2 * real[loop2 + i1];
				b2 += z1 * imag[loop2 + i1];

				real[loop2] = a1 + b1;
				imag[loop2] = a2 + b2;

				real[loop2 + i1] = a1 - b1;
				imag[loop2 + i1] = a2 - b2;
			}

			i3 += (i1 << 1);
			i4 += (i1 << 1);
		}

		i1 >>= 1;
		i2 <<= 1;
	}
}

/*
 *	Display the frequency domain.
 *
 *	The filters are aranged so that DC is in the middle filter.
 *	Thus -Doppler is on the left, +Doppler on the right.
 */

static void display()
{
	register int	c, loop;
	int		n, x;

	max_amp();
	n = samples / 2;

	for (loop = n; loop < samples; loop++)  {
		x = (int)(magnitude(loop) * (double)56.0 / maxn);
		printf("%d\t|", loop - n);
		c = 0;
		while (++c <= x)
			putchar('=');

		putchar('\n');
	}

	for (loop = 0; loop < n; loop++)  {
		x = (int)(magnitude(loop) * (double)56.0 / maxn);
		printf("%d\t|", loop + n);
		c = 0;
		while (++c <= x)
			putchar('=');

		putchar('\n');
	}
}

/*
 *	Find maximum amplitude
 */

static void max_amp()
{
	register int	loop;
	double		mag;

	maxn = (double)0.0;
	for (loop = 0; loop < samples; loop++)  {
		if ((mag = magnitude(loop)) > maxn)
			maxn = mag;
	}
}

/*
 *	Calculate Power Magnitude
 */

#ifdef SIMULATE
/*
 *	Simulate sqrt(i^2 + q^2).  Do this by adding half the smaller to
 *	the larger.  See also excellent description by Bob Leedom in the
 *	June 1979 Byte Magazine Technical Forum P. 188.  I originally saw
 *	The algorithm used in both an FAA Moving Target Detector (MTD) and
 *	a Westinghouse radar Envelope Detector.  Both in Hardware of course,
 *	where the binary shifting worked very fast.  Don't need high accuracy
 *	to tell you the storm is gonna remove you from Kansas (eh - Toto)...
 */

static double magnitude(n)
int	n;
{
	double	i, q;

	n = permute(n);

	i = fabs(real[n]);	/* mask the sign bit */
	q = fabs(imag[n]);

	if (i > q)
		return (i + q / (double)2.0);	/* shift right 1 and add */
	else
		return (q + i / (double)2.0);
}
#else

/*
 *	Do it the normal way
 */

static double magnitude(n)
int	n;
{
	double	i, q;

	n = permute(n);
	i = real[n] * real[n];
	q = imag[n] * imag[n];

	return sqrt(i + q);
}
#endif

/*
 *	Bit reverse the number
 *
 *	Change 11100000b to 00000111b or vice-versa
 */

static int permute(index)
int	index;
{
	register int	loop;
	unsigned	n1;
	int		result;

	n1 = samples;
	result = 0;

	for (loop = 0; loop < power; loop++)  {
		n1 >>= 1;
		if (index < n1)
			continue;

		/* Unix has a rounding error that MS-DOS compilers don't */
		/* so go ahead and add some bits (.05) */

		result += (int)((double).05 + pow((double)2.0, (double)loop));
		index -= n1;
	}

	return result;
}

#ifdef SPEED
/*
 *	Pre-compute the sine and cosine tables
 */

static void build_trig()
{
	register int	loop;
	double		angle, increment;

	angle     = (double)0.0;
	increment = TWO_PI * ((double)1.0 / (double)samples);

	for (loop = 0; loop < samples; loop++)  {
		sine[loop]   = sin(angle);
		cosine[loop] = cos(angle);
		angle       += increment;
	}
}
#endif

/* EOF */
